#include "stdint.h"
#include "unixtypes.h"
#include "stralloc.h"
#include "fmt.h"
#include "str.h"
#include "stdint.h"
#include "seek.h"
#include "error.h"
#include "open.h"
#include "unix.h"

#include "install_log.h"

void _exit(int);

static const char FATAL[] = "install_log: fatal: ";
static const char log_file_name[] = "install.log";
static const char blank[] = " ";
static const char slash[] = "/";
static const char eol[] = "\n";

static void puterr(const char *s)
{
  if (write(2, s, str_len(s)) < 0) _exit(111);
}

static void die1x(const char *x1)
{
  puterr(FATAL);
  puterr(x1);
  puterr("\n");
  _exit(111);
}

static void die_sys(const char *x1, const char *x2)
{
  const char *sys_error_str;
  sys_error_str = error_str(errno);

  puterr(FATAL);

  if (x1) puterr(x1);
  if (x2) puterr(x2);

  puterr(": ");
  puterr(sys_error_str);
  puterr("\n");
  _exit(111);
}

static void die1sys(const char *x1)
{
  die_sys(x1, (const char *)0);
}

static void die2sys(const char *x1, const char *x2)
{
  die_sys(x1, x2);
}

static void die_nomem(void)
{
  die1x("out of memory");
}

static void close_log(int fdout)
{
  if (fsync(fdout) < 0) {
    die2sys("unable to write ", log_file_name);
  }
  if (close(fdout) < 0) { /* NFS silliness */
    die2sys("unable to write ", log_file_name);
  }
}

static int fdsourcedir = -1;

void install_log_init(int fd)
{
  int fdout;
  fdsourcedir = fd;
  if (fchdir(fdsourcedir) < 0) {
    die1sys("unable to switch back to source directory");
  }
  fdout = open_trunc(log_file_name);
  if (fdout < 0) {
    die2sys("unable to create ", log_file_name);
  }
  close_log(fdout);
  if (chmod(log_file_name, 0666) < 0) {
    die2sys("unable to chmod ", log_file_name);
  }
}

static stralloc buffer = STRALLOC;
static char num_buff[FMT_ULONG+1];

static void buff_cat_int(int n)
{
  unsigned int i;
  unsigned int offset;

  offset = fmt_minus(num_buff, n);
  if (offset) {
    n = -n;
  }
  i = (unsigned int)n;
  num_buff[fmt_uint(num_buff + offset, i) + offset] = '\0';
  if (!stralloc_cats(&buffer, num_buff)) die_nomem();
}

static stralloc path = STRALLOC;

static void make_path(const char *p1, const char *p2)
{
  if (!stralloc_copys(&path, p1)) die_nomem();
  stralloc_trim(&path, "/", 1);
  if (!stralloc_append(&path, slash)) die_nomem();
  if (p2[0] == '/') {
    stralloc_trim(&path, "/", 1);
  }
  if (!stralloc_cats(&path, p2)) die_nomem();
}

static void make_path3(const char *p1, const char *p2, const char *p3)
{
  if (!stralloc_copys(&path, p1)) die_nomem();
  stralloc_trim(&path, "/", 1);
  if (!stralloc_append(&path, slash)) die_nomem();
  if (p2[0] == '/') {
    stralloc_trim(&path, "/", 1);
  }
  if (!stralloc_cats(&path, p2)) die_nomem();
  stralloc_trim(&path, "/", 1);
  if (!stralloc_append(&path, slash)) die_nomem();
  if (p3[0] == '/') {
    stralloc_trim(&path, "/", 1);
  }
  if (!stralloc_cats(&path, p3)) die_nomem();
}

static void log_path_and_save(uid_t uid, gid_t gid, mode_t mode)
{
  int fdout;

  if (!stralloc_append(&buffer, blank)) die_nomem();

  if (!stralloc_cat(&buffer, &path)) die_nomem();
  if (!stralloc_append(&buffer, blank)) die_nomem();

  buff_cat_int((int)uid);
  if (!stralloc_append(&buffer, blank)) die_nomem();

  buff_cat_int((int)gid);
  if (!stralloc_append(&buffer, blank)) die_nomem();

  num_buff[fmt_8long(num_buff, mode)] = '\0';
  if (!stralloc_cats(&buffer, num_buff)) die_nomem();

  if (!stralloc_append(&buffer, eol)) die_nomem();

  if (fchdir(fdsourcedir) < 0) {
    die1sys("unable to switch back to source directory");
  }
  fdout = open_write(log_file_name);
  if (fdout < 0) {
    die2sys("unable to open ", log_file_name);
  }
  if (seek_end(fdout) < 0) {
    close(fdout);
    die2sys("unable to seek to end ", log_file_name);
  }
  if (write(fdout, buffer.s, buffer.len) != (int)buffer.len) {
    close(fdout);
    die2sys("unable to write ", log_file_name);
  }
  close_log(fdout);
}

static void log_mkdir(uid_t uid, gid_t gid, mode_t mode)
{
  if (!stralloc_copys(&buffer, "mkdir")) die_nomem();
  log_path_and_save(uid, gid, mode);
}

void log_h(const char *home, uid_t uid, gid_t gid, mode_t mode)
{
  if (!stralloc_copys(&path, home)) die_nomem();
  log_mkdir(uid, gid, mode);
}

void log_d(const char *home, const char *subdir, uid_t uid, gid_t gid, mode_t mode)
{
  make_path(home, subdir);
  log_mkdir(uid, gid, mode);
}

void log_c(const char *home, const char *subdir, const char *file, uid_t uid, gid_t gid, mode_t mode)
{
  if (!stralloc_copys(&buffer, "copy")) die_nomem();
  if (!stralloc_append(&buffer, blank)) die_nomem();
  if (!stralloc_cats(&buffer, file)) die_nomem();

  make_path(home, subdir);
  log_path_and_save(uid, gid, mode);
}

void log_p(const char *home, const char *fifo, uid_t uid, gid_t gid, mode_t mode)
{
  if (!stralloc_copys(&buffer, "fifo")) die_nomem();
  make_path(home, fifo);
  log_path_and_save(uid, gid, mode);
}

void log_z(const char *home, const char *subdir, const char *file, unsigned int len, uid_t uid, gid_t gid, mode_t mode)
{
  if (!stralloc_copys(&buffer, "zero")) die_nomem();
  if (!stralloc_append(&buffer, blank)) die_nomem();

  num_buff[fmt_uint(num_buff, len)] = '\0';
  if (!stralloc_cats(&buffer, num_buff)) die_nomem();

  make_path3(home, subdir, file);
  log_path_and_save(uid, gid, mode);
}
