#include "stdint.h"
#include "unixtypes.h"
#include "unix.h"
#include "str.h"
#include "ip4.h"
#include "ip6.h"

void _exit(int);

static unsigned int errors = 0;

static void outs(const char *s)
{
  while (*s) {
    if (write(1, s, 1) < 0) _exit(111);
    s++;
  }
}

static void err_compare(const char *input, const char *expected, const char *output)
{
  errors++;
  outs("input='");
  outs(input);
  outs("'\n");
  outs("output='");
  outs(output);
  outs("'\n");
  outs("expected='");
  outs(expected);
  outs("'\n");
}

static void err_syntax(const char *input)
{
  errors++;
  outs("input='");
  outs(input);
  outs("' - unexpected syntax error\n");
}

static void err_no_syntax(const char *input)
{
  errors++;
  outs("input='");
  outs(input);
  outs("' - syntax error not detected\n");
}

static char buffer[IP6_FMT];

static void test_scan(const char *input, const char *expected, unsigned int test)
{
  ip6_address ip;

  if (!ip6_scan(&ip, input)) {
    if (test) err_syntax(input);
    return;
  }
  if (test) {
    buffer[ip6_fmt_uncompressed(&ip, buffer)] = '\0';
    if (str_diff(buffer, expected)) err_compare(input, expected, buffer);
    return;
  }
  err_no_syntax(input);
}

static void test_scanbracket(const char *input, const char *expected, unsigned int test)
{
  ip6_address ip;

  if (!ip6_scanbracket(&ip, input)) {
    if (test) err_syntax(input);
    return;
  }
  if (test) {
    buffer[ip6_fmt_uncompressed(&ip, buffer)] = '\0';
    if (str_diff(buffer, expected)) err_compare(input, expected, buffer);
    return;
  }
  err_no_syntax(input);
}

int main(int argc, char **argv)
{
  if (argc != 1) _exit(111);  /* keep compiler quiet */
  if (!*argv[0]) _exit(111);  /* keep compiler quiet */
/*
  syntax = good
*/
  test_scan("1:2:3:4:5:6:7:8", "1:2:3:4:5:6:7:8", 1);

  test_scan("::", "0:0:0:0:0:0:0:0", 1);
  test_scan("::1", "0:0:0:0:0:0:0:1", 1);
  test_scan("1::", "1:0:0:0:0:0:0:0", 1);

  test_scan("1234:5678::", "1234:5678:0:0:0:0:0:0", 1);
  test_scan("1234:5678::9", "1234:5678:0:0:0:0:0:9", 1);

  test_scan("1234:5678::0.1.0.2", "1234:5678:0:0:0:0:1:2", 1);
  test_scan("1:2:3:4:5:6:0.7.0.8", "1:2:3:4:5:6:7:8", 1);

  test_scan("1:2:3:4:5:6:7:8%9", "1:2:3:4:5:6:7:8%9", 1);

  test_scanbracket("[1:2:3:4:5:6:7:8]", "1:2:3:4:5:6:7:8", 1);
  test_scanbracket("[1:2:3:4:5:6:7:8%9]", "1:2:3:4:5:6:7:8%9", 1);
/*
  syntax = bad
*/
  test_scan("1:2", "", 0);
  test_scan("1:2:3:4:5:6:7:89abc", "", 0);
  test_scan("1234:5678::9::", "", 0);
  test_scan("12345:6789::", "", 0);
  test_scan("::12345:6789", "", 0);
  test_scan("1:2:3:4:5:6:7:8:9", "", 0);
  test_scan("1:2:3:4:5:6:7:8%", "", 0);
  test_scan("1234:5678::1.0.2", "", 0);
  test_scan("1234:5678:1.0.2::9", "", 0);
  test_scan("1234:5678::1.0.2:9", "", 0);
  test_scan("1:2:3:4:5:6:0.7.0.8:9", "", 0);

  if (errors) {
    outs("test_ip6 failed tests!\n");
    _exit(2);
  }

  _exit(0);
}
