#include "stdint.h"
#include "unixtypes.h"
#include "unix.h"
#include "stralloc.h"
#include "error.h"

void _exit(int);

static unsigned int errors = 0;

static void outs(const char *s)
{
  while (*s) {
    if (write(1, s, 1) < 0) _exit(111);
    s++;
  }
}
static void outsa(const stralloc *sa)
{
  if (write(1, sa->s, sa->len) < 0) _exit(1);
}

static void die_nomem(void)
{
  outs("fatal: out of memory\n");
  _exit(1);
}

static void err_compare(const char *input, const char *expected, const stralloc *output)
{
  errors++;
  outs("input='");
  outs(input);
  outs("'\n");
  outs("output='");
  outsa(output);
  outs("'\n");
  outs("expected='");
  outs(expected);
  outs("'\n");
}

static void err_syntax(const char *input)
{
  errors++;
  outs("input='");
  outs(input);
  outs("' - unexpected syntax error\n");
}

static void err_no_syntax(const char *input)
{
  errors++;
  outs("input='");
  outs(input);
  outs("' - syntax error not detected\n");
}

static stralloc buffer = STRALLOC;
static stralloc output = STRALLOC;

static void test_encode(const char *input, const char *expected)
{
  if (!stralloc_copys(&buffer, input)) die_nomem();
  if (!stralloc_hex_encode(&output, &buffer)) die_nomem();
  if (stralloc_case_diffs(&output, expected)) err_compare(input, expected, &output);
}

static void test_decode(const char *input, const char *expected, unsigned int test)
{
  if (!stralloc_copys(&buffer, input)) die_nomem();
  if (!stralloc_hex_decode(&output, &buffer)) {
    if (errno == error_nomem) die_nomem();
    if (test) err_syntax(input);
    return;
  }
  if (test) {
   if (stralloc_diffs(&output, expected)) err_compare(input, expected, &output);
   return;
  }
  err_no_syntax(input);
}

int main(int argc, char **argv)
{
  if (argc != 1) _exit(111);  /* keep compiler quiet */
  if (!*argv[0]) _exit(111);  /* keep compiler quiet */

  test_encode("", "");
  test_encode("f", "66");
  test_encode("fo", "666F");
  test_encode("foo", "666F6F");
  test_encode("foob", "666F6F62");
  test_encode("fooba", "666F6F6261");
  test_encode("foobar", "666F6F626172");

  test_decode("", "", 1);
  test_decode("66", "f", 1);
  test_decode("666F", "fo", 1);
  test_decode("666F6F", "foo", 1);
  test_decode("666F6F62", "foob", 1);
  test_decode("666F6F6261", "fooba", 1);
  test_decode("666F6F626172", "foobar", 1);
  test_decode("666F6F6261720", "", 0);  /* test uneven */
  test_decode("@666F6F6261720", "", 0);  /* test bad char */

  if (errors) {
    outs("test_stralloc_hex failed tests!\n");
    _exit(2);
  }

  _exit(0);
}
