#include "stdint.h"
#include "unixtypes.h"
#include "unix.h"
#include "stralloc.h"
#include "error.h"

void _exit(int);

static unsigned int errors = 0;

static void outs(const char *s)
{
  while (*s) {
    if (write(1, s, 1) < 0) _exit(111);
    s++;
  }
}
static void outsa(const stralloc *sa)
{
  if (write(1, sa->s, sa->len) < 0) _exit(1);
}

static void die_nomem(void)
{
  outs("fatal: out of memory\n");
  _exit(1);
}

static void err_compare(const char *input, const char *expected, const stralloc *output)
{
  errors++;
  outs("input='");
  outs(input);
  outs("'\n");
  outs("output='");
  outsa(output);
  outs("'\n");
  outs("expected='");
  outs(expected);
  outs("'\n");
}

static void err_syntax(const char *input)
{
  errors++;
  outs("input='");
  outs(input);
  outs("' - unexpected syntax error\n");
}

static void err_no_syntax(const char *input)
{
  errors++;
  outs("input='");
  outs(input);
  outs("' - syntax error not detected\n");
}

static stralloc buffer = STRALLOC;
static stralloc output = STRALLOC;

static void test_encode(const char *input, const char *expected)
{
  if (!stralloc_copys(&buffer, input)) die_nomem();
  if (!stralloc_qp_encode(&output, &buffer)) die_nomem();
  if (stralloc_diffs(&output, expected)) err_compare(input, expected, &output);
}

static void test_decode(const char *input, const char *expected, unsigned int test)
{
  if (!stralloc_copys(&buffer, input)) die_nomem();
  if (!stralloc_qp_decode(&output, &buffer)) {
    if (errno == error_nomem) die_nomem();
    if (test) err_syntax(input);
    return;
  }
  if (test) {
   if (stralloc_diffs(&output, expected)) err_compare(input, expected, &output);
   return;
  }
  err_no_syntax(input);
}

int main(int argc, char **argv)
{
  if (argc != 1) _exit(111);  /* keep compiler quiet */
  if (!*argv[0]) _exit(111);  /* keep compiler quiet */

  test_encode("", "");
  test_encode("f", "f");
  test_encode("fo", "fo");
  test_encode("foo", "foo");
  test_encode("foo=", "foo=3D");
  test_encode("foo=b", "foo=3Db");
  test_encode("foo=ba", "foo=3Dba");
  test_encode("foo=bar", "foo=3Dbar");
  test_encode("foo=bar\n", "foo=3Dbar=0A");

  test_decode("", "", 1);
  test_decode("f", "f", 1);
  test_decode("fo", "fo", 1);
  test_decode("foo", "foo", 1);
  test_decode("foo=3D", "foo=", 1);
  test_decode("foo=3Db", "foo=b", 1);
  test_decode("foo=3Dba", "foo=ba", 1);
  test_decode("foo=3Dbar", "foo=bar", 1);
  test_decode("foo=3Dbar=0A", "foo=bar\n", 1);
  test_decode("foo=GZbar", "", 0);  /* bad hex */
  test_decode("foo\251bar\n", "", 0);  /* non printable not allowed */

  if (errors) {
    outs("test_stralloc_qp failed tests!\n");
    _exit(2);
  }

  _exit(0);
}
